#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Copyright (C) 2008 Thomas Holder, http://sf.net/users/speleo3/
Distributed under the terms of the GNU General Public License v2

Converts Survex 3D files (*.3d) to SVG.
Usage: python 3dtosvg.py [OPTIONS] FILE.3d

  --scale=ARG        Import with a scale of 1:ARG (default 100)
  --view=[0,1,2]     0: Plan (default), 1: Profile, 2: Extend
  --bearing=[0-359]  Bearing in degrees north (default 0)
  --markers=[0-2]    0: No station markers, 1: Display stations as small
                     circles (default), 2: ditto as triangles
  --dpi=ARG          Resolution in DPI (default 90)
  --extend-cmd=ARG   The "extend" program is part of Survex and required
                     for --view=2. If it is not found inside PATH, you may
                     specify the absolute path for the binary with ARG.
                     Example: --extend-cmd="C:\Programme\Survex\extend.exe"
                     Example: --extend-cmd="/usr/local/bin/extend"

If you select extend view and there is a *.espec file with same
basename as the *.3d file, it will be passed to the extend binary
with the --specfile option. See "extend" manpage for details.

Related:
 * http://survex.com/
 * http://trac.survex.com/browser/trunk/doc/3dformat.htm
'''

import sys, math, os

args = {
	'scale': 100,
	'view': 0,
	'bearing': 0,
	'markers': 1,
	'dpi': 90,
	'extend-cmd': 'extend',
}

infile = ''

def die(msg):
	sys.stderr.write(msg + "\n")
	sys.exit(1)

sys.argv.pop(0)
for arg in sys.argv:
	if arg.startswith('--'):
		pair = arg[2:].split('=', 2)
		if pair[1].isdigit():
			pair[1] = int(pair[1])
		elif pair[1] == 'true':
			pair[1] = 1
		elif pair[1] == 'false':
			pair[1] = 0
		args[pair[0]] = pair[1]
	else:
		infile = arg

if len(infile) == 0:
	die("no filename given")

if args['view'] == 2:
	if not infile.endswith('_extend.3d'):
		command = args['extend-cmd'] + ' "' + infile + '"'
		if os.path.exists(infile[:-3] + '.espec'):
			command += ' --specfile="' + infile[:-3] + '.espec"'
		pipe = os.popen(command)
		if pipe.close() != None:
			die("extend failed")
		infile = infile[:-3] + '_extend.3d'
	args['bearing'] = 0

f = open(infile, 'rb')

line = f.readline() # File ID
if not line.startswith('Survex 3D Image File'):
	die("not a Survex 3D File, aborting")

line = f.readline() # File format version
line = unicode(f.readline(), 'latin1') # Survex title
if line.rstrip().endswith(' (extended)'):
	args['view'] = 2
	args['bearing'] = 0

line = f.readline() # Timestamp

_cos = math.cos(math.radians(args['bearing']))
_sin = math.sin(math.radians(args['bearing']))

coords = []

def read_xyz():
	x = read_int(4, 1)
	y = read_int(4, 1)
	z = read_int(4, 1)
	if args['bearing']:
		global _cos, _sin
		[ x, y ] = [ x * _cos - y * _sin, x * _sin + _cos * y ]
	if args['view'] > 0:
		[ y, z ] = [ z, y ]
	return [ x, y, z ]

def read_int(len, sign):
	int = 0
	for i in range(len):
		int |= read_byte() << (8 * i)
	if sign and (int >> (8 * len - 1)):
		int -= (1 << 8 * len)
	return int

def read_len():
	len = read_byte()
	if len == 0xfe:
		len += read_int(2, 0)
	elif len == 0xff:
		len = read_int(4, 0)
	return len

def skip_label():
	len = read_len()
	skip_bytes(len)

def skip_bytes(n):
	# note: f.seek(n) failes on windows in some weird cases!
	f.read(n)

def read_byte():
	byte = f.read(1)
	if len(byte) != 1:
		return -1
	return ord(byte)

while 1:
	byte = read_byte()
	if byte == -1:
		break
	
	if byte == 0x00:
		# STOP
		continue
	elif byte <= 0x0e:
		# TRIM
		continue
	elif byte <= 0x0f:
		# MOVE
		xyz = read_xyz()
		coords.append('M')
		coords.extend(xyz)
	elif byte <= 0x1f:
		# TRIM
		continue
	elif byte <= 0x20:
		# DATE
		skip_bytes(4)
	elif byte <= 0x21:
		# DATE
		skip_bytes(8)
	elif byte <= 0x22:
		# Error info
		skip_bytes(5 * 4)
	elif byte <= 0x2f:
		# Reserved
		continue
	elif byte <= 0x31:
		# XSECT
		skip_label()
		skip_bytes(4 * 2)
	elif byte <= 0x33:
		# XSECT
		skip_label()
		skip_bytes(4 * 4)
	elif byte <= 0x3f:
		# Reserved
		continue
	elif byte <= 0x7f:
		# LABEL
		skip_label()
		read_xyz()
	elif byte <= 0xbf:
		# LINE
		skip_label()
		xyz = read_xyz()
		coords.append('L')
		coords.extend(xyz)
	elif byte <= 0xff:
		# Reserved
		continue

# find minimum
min_x = coords[1]
min_y = coords[2]
for i in range(4, len(coords), 4):
	if (min_x > coords[i + 1]):
		min_x = coords[i + 1]
	if (min_y > coords[i + 2]):
		min_y = coords[i + 2]

scale = args['dpi'] / 2.54 / args['scale']
stroke = args['scale'] / 50
marker = [
	'none',
	'url(#StationCircle)',
	'url(#StationTriangle)',
][args['markers']]

print """<?xml version="1.0" encoding="UTF-8"?>
<svg xmlns="http://www.w3.org/2000/svg"
	xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"
	xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape">
<defs>
	<marker style="overflow:visible" id="StationCircle"
					   inkscape:stockid="StationCircle">
		<circle r="2"
			style="fill:none;stroke:#666;stroke-width:0.7" />
	</marker>
	<marker style="overflow:visible" id="StationTriangle"
					   inkscape:stockid="StationTriangle">
		<path d="M -4,2 L 4,2 L 0,-4.5 Z"
			style="fill:none;stroke:#666;stroke-width:1" />
	</marker>
</defs>
<path
	style="fill:none;stroke:#900;stroke-width:%d;stroke-linecap:round;stroke-linejoin:round;marker-start:%s;marker-mid:%s;marker-end:%s"
	transform="scale(%f,%f)" """ % \
	(
		stroke,
		marker, marker, marker,
		scale, scale
	)

sys.stdout.write('d="')
for i in range(0, len(coords), 4):
	sys.stdout.write(coords[i] + " " +
		str(  coords[i + 1] - min_x) + "," +
		str(- coords[i + 2] + min_y) + " ")
sys.stdout.write('"');

print """ />
</svg>"""
